# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import logging
import os
import signal
import sys
from typing import Optional

from logger.config import ConfigOverrideOption, get_config
from logger.constants import Constants, ExitCode
from logger.error import TerminateException, raise_terminate_exception
from logger.log import LoggingService
from logger.measure.controller import Controller
from logger.process import prepare_process
from logger.utils.env import parse_int
from logger.utils.lock import ExclusiveLock, LockError


def main():  # noqa: C901
    """メイン関数"""

    # プロセス初期化
    # - シグナルハンドラー登録 : SIGTERM, SIGINT 発生時は例外処理にて対処する
    signal.signal(signal.SIGTERM, raise_terminate_exception)
    signal.signal(signal.SIGINT, raise_terminate_exception)

    # - サブプロセス生成方式設定
    prepare_process()

    # - ロギング機構取得
    logging_service = LoggingService.get_instance()

    # - 排他ロック生成
    lock = ExclusiveLock("/tmp/measure.lock")

    # メイン処理開始
    controller: Optional[Controller] = None
    try:
        # ロギング開始
        logging_service.start()
        logger = logging.getLogger(__name__)
        # - プロセスIDを出力
        logger.debug(f"process id: {os.getpid()}")

        # 多重起動防止
        try:
            lock.acquire()
        except LockError:
            logger.error("Another measurement is running, can't start measurement")
            return ExitCode.EX_NG

        # 引数処理
        args = sys.argv
        # - 引数は必ず2個（1番目はプログラム名）
        if len(args) != 2:
            logger.error(
                'Invalid arguments length. The argument must be "auto", "manual", or an integer.'
            )
            return ExitCode.EX_NG

        # - "auto", "manual"
        if args[1] in [Constants.START_TYPE_AUTO, Constants.START_TYPE_MANUAL]:
            start_type = args[1]
            meas_time_sec_arg = None
        # - "schedule"
        else:
            start_type = Constants.START_TYPE_SCHEDULE
            meas_time_sec_arg = parse_int(args[1])
            if meas_time_sec_arg is None:
                logger.error("Measurement time argument must be an integer")
                return ExitCode.EX_NG

        # 設定オブジェクト取得
        config = get_config(ConfigOverrideOption(measure_time_sec=meas_time_sec_arg))
        if config is None:
            logger.error("Config file has wrong value, can't start measurement")
            return ExitCode.EX_NG

        # 計測コントローラ生成
        controller = Controller(config=config, start_type=start_type)
        try:
            # 計測初期化
            controller.init()
        except TerminateException:
            # シグナルにより中断された場合、正常終了とする
            return ExitCode.EX_OK
        except Exception:
            logger.exception("Failed to initialize measurement")
            logger.error("Can't start measurement")
            return ExitCode.EX_NG

        try:
            # 計測開始
            controller.start()
        except TerminateException:
            # シグナルにより中断された場合、正常終了とする
            # - この後の finally ブロックで計測終了が処理される
            # - 各所でログが出力されるため、ここではログ出力しない
            return ExitCode.EX_OK
        except Exception:
            logger.exception("Failed to measure")
            logger.error("Can't continue measurement")
            return ExitCode.EX_NG

        # 計測成功
        return ExitCode.EX_OK

    except TerminateException:
        # 上の try ブロック外でのシグナル中断も正常終了とする
        return ExitCode.EX_OK
    except Exception:
        logger.exception("Unknown error occurred")
        return ExitCode.EX_NG

    finally:
        # 計測終了〜プロセス終了
        # - 計測コントローラ停止
        output_dir: Optional[str] = None
        if controller:
            output_dir = controller.get_output_dir()
            # 計測停止指示: terminate=True として停止を指示する
            # - タイマー実行で計測終了している場合は、停止指示は無視される
            controller.stop(terminate=True)

        # - ロギング停止
        logging_service.close(output_dir)

        # - 排他ロック解放
        lock.release()


if __name__ == "__main__":
    code = main()
    sys.exit(code)
