# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

from dataclasses import dataclass
from typing import List, Optional, TypedDict, Union

from logger.constants import ConfigKeys, Constants
from logger.utils.env import get_bool, get_float, get_int, get_str
from logger.utils.validator import Validator


@dataclass
class Config:
    """設定ファイルで指定された設定値を保持するクラス"""

    logger_id: str
    output_path: str
    initial_wait: int
    measure_time_sec: int
    file_rotate_min: int
    baud_rate: int
    A342_physical: str
    A342_mode: str
    A342_rmspp_output_interval: int
    A352_sps: int
    sensor_data_diag: bool
    sensor_data_diag_sec: float


class ConfigOverrideOption(TypedDict, total=False):
    """設定ファイルの設定値を上書きするためのクラス"""

    measure_time_sec: Optional[int]


def get_config(option: Optional[ConfigOverrideOption] = None) -> Union[Config, None]:
    """
    環境変数(設定ファイル)から設定値を読み込み設定オブジェクトを返す
    - 設定値の読み込みにあたり、設定値が想定された形式であるかバリデーションを行う
        - すべてのバリデーションに成功した場合、設定オブジェクトが返る
    - option による設定値の上書きが可能

    Args:
        option (ConfigOverrideOption): 環境変数の設定値を上書きする値を保持するオブジェクト
            - 現在は計測秒数のみ対応
    Returns:
        Union[Config,None]: 設定オブジェクト
            - バリデーションに失敗した場合はNoneが返る
    """
    validator = Validator()
    results: List[bool] = []

    # LOGGER_ID
    logger_id = get_str(ConfigKeys.LOGGER_ID)
    results.append(
        validator.match(logger_id, ConfigKeys.LOGGER_ID, Constants.LOGGER_ID_PATTERN)
    )

    # OUTPUT_PATH
    output_path = get_str(ConfigKeys.OUTPUT_PATH)
    results.append(validator.exist_dir(output_path, ConfigKeys.OUTPUT_PATH))

    # INITIAL_WAIT
    initial_wait = get_int(ConfigKeys.INITIAL_WAIT)
    results.append(
        validator.between(
            initial_wait,
            ConfigKeys.INITIAL_WAIT,
            Constants.INITIAL_WAIT_MIN,
            Constants.INITIAL_WAIT_MAX,
        )
    )

    # MEASURE_TIME_SEC
    # - option による上書きを処理する
    measure_time_sec = (
        get_int(ConfigKeys.MEASURE_TIME_SEC)
        if option is None or option.get("measure_time_sec") is None
        else option.get("measure_time_sec")
    )
    results.append(
        validator.between(
            measure_time_sec,
            ConfigKeys.MEASURE_TIME_SEC,
            Constants.MEASURE_TIME_SEC_MIN,
            Constants.MEASURE_TIME_SEC_MAX,
        )
    )

    # FILE_ROTATE_MIN
    file_rotate_min = get_int(ConfigKeys.FILE_ROTATE_MIN)
    results.append(
        validator.between(
            file_rotate_min,
            ConfigKeys.FILE_ROTATE_MIN,
            Constants.FILE_ROTATE_MIN_MIN,
            Constants.FILE_ROTATE_MIN_MAX,
        )
    )

    # BAUD_RATE
    baud_rate = get_int(ConfigKeys.BAUD_RATE)
    results.append(
        validator.one_of(baud_rate, ConfigKeys.BAUD_RATE, Constants.BAUD_RATES)
    )

    # A342_PHYSICAL
    A342_physical = get_str(ConfigKeys.A342_PHYSICAL)
    results.append(
        validator.one_of(
            A342_physical, ConfigKeys.A342_PHYSICAL, Constants.A342_PHYSICALS
        )
    )

    # A342_MODE
    A342_mode = get_str(ConfigKeys.A342_MODE)
    results.append(
        validator.one_of(A342_mode, ConfigKeys.A342_MODE, Constants.A342_MODES)
    )

    # A342_RMSPP_OUTPUT_INTERVAL
    A342_rmspp_output_interval = get_int(ConfigKeys.A342_RMSPP_OUTPUT_INTERVAL)
    results.append(
        validator.between(
            A342_rmspp_output_interval,
            ConfigKeys.A342_RMSPP_OUTPUT_INTERVAL,
            Constants.A342_RMS_PP_OUTPUT_INTERVAL_MIN,
            Constants.A342_RMS_PP_OUTPUT_INTERVAL_MAX,
        )
    )

    # A352_SPS
    A352_sps = get_int(ConfigKeys.A352_SPS)
    results.append(validator.one_of(A352_sps, ConfigKeys.A352_SPS, Constants.A352_SPSS))

    # SENSOR_DATA_DIAG
    sensor_data_diag = get_bool(ConfigKeys.SENSOR_DATA_DIAG)
    results.append(
        validator.one_of(sensor_data_diag, ConfigKeys.SENSOR_DATA_DIAG, [True, False])
    )

    # SENSOR_DATA_DIAG_SEC
    sensor_data_diag_sec = get_float(ConfigKeys.SENSOR_DATA_DIAG_SEC)
    results.append(
        validator.greater_than(sensor_data_diag_sec, ConfigKeys.SENSOR_DATA_DIAG_SEC, 0)
    )
    results.append(
        validator.lesser_than_or_equal_to(
            sensor_data_diag_sec, ConfigKeys.SENSOR_DATA_DIAG_SEC, 10
        )
    )

    # 結果判定： いずれか失敗している場合は失敗
    if not all(results):
        return None

    # 設定オブジェクト構築・返却
    # - ここまでの処理で適正な値を保持していることが確認できているため、型チェックを無効化
    return Config(
        logger_id=logger_id,  # type: ignore
        output_path=output_path,  # type: ignore
        initial_wait=initial_wait,  # type: ignore
        measure_time_sec=measure_time_sec,  # type: ignore
        file_rotate_min=file_rotate_min,  # type: ignore
        baud_rate=baud_rate,  # type: ignore
        A342_physical=A342_physical,  # type: ignore
        A342_mode=A342_mode,  # type: ignore
        A342_rmspp_output_interval=A342_rmspp_output_interval,  # type: ignore
        A352_sps=A352_sps,  # type: ignore
        sensor_data_diag=sensor_data_diag,  # type: ignore
        sensor_data_diag_sec=sensor_data_diag_sec,  # type: ignore
    )
