# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import datetime
import logging
import logging.handlers
import multiprocessing
import os
import shutil
import signal
import sys
import traceback
from typing import Callable, Optional

from logger.constants import ConfigKeys, Constants
from logger.utils.dir import root_dir
from logger.utils.env import get_str


class LoggingClient:
    """プロセスセーフなロギング機構をセットアップするために、サブプロセスに受け渡すクラス

    Attributes:
        queue (multiprocessing.Queue): リスナープロセスにログメッセージを送出するqueue
        configure (Callable[[multiprocessing.Queue], Any]): プロセスで使用されるロガーにqueueハンドラーをセットアップする関数
    """

    def __init__(
        self,
        queue: multiprocessing.Queue,
        configure: Callable[[multiprocessing.Queue], None],
    ) -> None:
        self.queue = queue
        self.configure = configure


class LoggingService:
    """プロセスセーフなロギングを提供するクラス

    Python 公式の [Logging クックブック](https://docs.python.org/ja/3/howto/logging-cookbook.html#logging-to-a-single-file-from-multiple-processes)
    に例示されている「複数プロセスから単一ファイルへのログ記録」の構成に従い、
    マルチプロセス環境でも安全にファイル出力可能なロギング機構を構築する。

    LoggingService はシングルトンとして構成し、
    ロギング機構全体の開始、出力プロセスの構築、サブプロセス側の構築、終了を管理する。
    """

    # ログレベル： 設定値から取得
    LOG_LEVEL: str = str(
        get_str(ConfigKeys.LOG_LEVEL)
        if get_str(ConfigKeys.LOG_LEVEL) in Constants.LOG_LEVELS
        else Constants.LOG_LEVEL_DEFAULT
    )

    # シングルトンインスタンス
    _instance: Optional["LoggingService"] = None

    # - シングルトンを成立させるため、通常のインスタンス生成を抑止
    def __new__(cls):
        raise NotImplementedError("Cannot Generate Instance By Constructor")

    @classmethod
    def __internal_new__(cls):
        instance = super().__new__(cls)
        instance.__init__()  # type: ignore
        return instance

    @classmethod
    def get_instance(cls) -> "LoggingService":
        """シングルトンの LoggingService インスタンスを取得するメソッド

        Returns:
            LoggingService:
        """
        if not cls._instance:
            cls._instance = cls.__internal_new__()
        return cls._instance  # type: ignore

    def __init__(self) -> None:
        self._listener_process: Optional[multiprocessing.Process] = None
        self._queue: multiprocessing.Queue = multiprocessing.Queue(-1)
        self._log_file: Optional[str] = None

    @staticmethod
    def _listen(queue: multiprocessing.Queue, log_file: str):
        """ロギングのキューを監視し、標準出力とログファイルへの出力を実行する関数

        LoggingService.start() 内で生成される出力用プロセスに登録されるため staticmethod
        """
        # シグナル処理： SIGINT は無視する
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        # ロギング構築
        root = logging.getLogger()
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(_Formatter())
        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(_Formatter())
        root.addHandler(file_handler)
        root.addHandler(stream_handler)

        # ロギング出力
        while True:
            try:
                record = queue.get()
                if record is None:
                    # close() 時に None が送信されるため終了
                    break

                logger = logging.getLogger(record.name)
                logger.handle(record)

            except Exception:
                print("Whoops! Problem:", file=sys.stderr)
                traceback.print_exc(file=sys.stderr)

    @staticmethod
    def _configure(queue: multiprocessing.Queue) -> None:
        """各プロセス上で動作し、キューにデータを登録するロギング機構を構築する"""
        h = logging.handlers.QueueHandler(queue)
        root = logging.getLogger()
        root.addHandler(h)
        root.setLevel(LoggingService.LOG_LEVEL)

    def start(self):
        """ログ出力プロセスを起動し、ロギング機構を開始する"""
        # 初期化済み抑止
        if self._listener_process:
            return

        # メインプロセスからのログ出力用に _configure 実行
        LoggingService._configure(self._queue)

        # 出力ファイル： 計測終了までは一時ファイルに出力する
        self._log_file = "{}/tmp_log/{}_measure.log".format(
            root_dir(), format(datetime.datetime.now(), "%Y%m%d_%H%M%S")
        )

        # ログ出力プロセス構築・開始
        self._listener_process = multiprocessing.Process(
            target=LoggingService._listen, args=(self._queue, self._log_file)
        )
        self._listener_process.start()

    def get_logging_client(self) -> LoggingClient:
        """サブプロセスでロギング機構を構築するためのオブジェクトを返す

        Returns:
            LoggingClient: サブプロセス用オブジェクト

        Raises:
            AssertionError: start メソッドが実行される前に呼び出した場合に発生
        """
        assert self._listener_process
        return LoggingClient(queue=self._queue, configure=LoggingService._configure)

    def close(self, output_dir: Optional[str] = None) -> None:
        """ロギング機構全体を終了し、ログファイルを計測フォルダに移動する

        Args:
            output_dir (Optional[str], optional): ログファイルの移動先ディレクトリ
                - 指定した場合、「指定ディレクトリ/measure.log」にログファイルを移動する
        """
        try:
            # ロギング出力プロセス終了
            if self._listener_process is not None:
                # - None を送信してプロセスを終了させる
                self._queue.put_nowait(None)
                self._listener_process.join()
                self._listener_process = None

            # ログファイル移動（計測完了時）
            if output_dir and self._log_file and os.path.isdir(output_dir):
                shutil.move(self._log_file, f"{output_dir}/measure.log")
                self._log_file = None

        except Exception as e:
            # この処理は main の終了時に呼ばれる
            # すでにログが閉じている可能性があるため、Error は無視して出力のみ
            print(f"logging service close failed with Error: {str(e)}")


class _Formatter(logging.Formatter):
    """ログ出力のフォーマッター"""

    def format(self, rec: logging.LogRecord) -> str:
        return "%-8s|%s %s: %s" % (
            rec.levelname,
            self.formatTime(rec),
            rec.name,
            rec.getMessage(),
        )


def init_stdout_logger(level: str) -> None:
    """標準出力のみにログ出力を行うロガーのセットアップ関数

    LoggingService によるマルチプロセス対応のロギング機構でなく、
    標準出力のみに出力する簡易なログを使いたいツール向け

    Args:
        level (str): ログレベル
    """
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(_Formatter())
    logging.basicConfig(handlers=[handler], level=level)
