# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import datetime
import logging
import math
import multiprocessing
import os
import platform
import threading
import time
from typing import List, Optional, Union

import serial.tools.list_ports

from logger.config import Config
from logger.constants import Constants, InfoKeys
from logger.error import LoggerError
from logger.measure.factory import LoggerFactory
from logger.measure.job import JobError
from logger.measure.sensor import Sensor, SensorDeviceError
from logger.utils.format import app_format
from logger.utils.timezone import get_time_zone


class Controller:
    """
    センサーによる計測をコントロールするクラス

    以下の処理を行う：
    - USB 接続されたセンサーを検索し、初期化する
    - タイミングに従い、センサーの計測を開始する
    - タイマーや外部イベントに従い、センサーの計測を停止する
    - センサーのセルフテストを実行する
    """

    def __init__(
        self, config: Config, start_type: str, factory: Optional[LoggerFactory] = None
    ) -> None:
        """
        Controller のコンストラクタ

        Args:
            config (Config): 設定情報を保持した Config オブジェクト
            start_type (str): main 開始時に指定された起動タイプ
            factory (Optional[LoggerFactory]): Logger の要素を生成するファクトリ
                - 省略時はデフォルトの LoggerFactory が適用される
        """
        self.config = config
        self._factory: LoggerFactory = (
            LoggerFactory(config) if factory is None else factory
        )
        self._start_type = start_type
        self._sensors: List[Sensor] = []
        self._stop_timer: Optional[threading.Timer] = None
        self._measuring: bool = False
        self._error_queue: Optional[multiprocessing.Queue] = None
        self._start_time: Optional[datetime.datetime] = None
        self._end_time: Optional[datetime.datetime] = None
        self._logger = logging.getLogger(__name__)

    def init(self) -> None:
        """計測の初期化を行う"""
        self._logger.info("Start initialization")

        # センサー検索： USBデバイスをサーチしてセンサを探す
        for port in Controller._list_comports():
            self._logger.debug(f"COM port found: {port}")

            # - センサーとの接続確立試行
            sensor = self._init_sensor(port)
            if sensor:
                self._sensors.append(sensor)

        # - センサーが見つからなければ終了
        if len(self._sensors) == 0:
            self._logger.error("No sensor found")
            raise LoggerError("No Ready Sensor found")

        # 接続できたセンサーに初期化を指示する
        for sensor in self._sensors:
            sensor.init()

    def start(self) -> None:  # noqa: C901
        """センサーによる計測を開始する

        Raises:
            self._error_queue.get: サブプロセスから発生したエラー
        """
        # 初回待機： スケジュール計測以外
        if self._start_type != Constants.START_TYPE_SCHEDULE:
            self._logger.info(f"Initial wait for {self.config.initial_wait} sec")
            time.sleep(self.config.initial_wait)

        # 計測開始
        # - 計測開始日時
        self._start_time = datetime.datetime.now()

        # - 出力先フォルダ
        out_dir: str = str(self.get_output_dir())
        os.makedirs(out_dir, exist_ok=True)

        # - 計測時間設定: 計測時間が設定されている場合
        if self.config.measure_time_sec > 0:
            # タイマー設定： タイマー発火時は stop() が呼ばれる
            self._stop_timer = threading.Timer(self.config.measure_time_sec, self.stop)
            self._stop_timer.start()

        # - 内部計測フラグ ON
        self._measuring = True

        # - 計測エラー通知用キュー
        self._error_queue = multiprocessing.Queue()

        # - センサーに計測開始指示
        self._logger.info("Start measurement")
        failed = set()
        for sensor in self._sensors:
            try:
                sensor.start(
                    error_queue=self._error_queue,
                    output_dir=out_dir,
                    factory=self._factory,
                )
            except Exception:
                self._logger.exception(f"Sensor: {sensor} failed to start measurement")
                failed.add(f"{sensor.model}_{sensor.serial}")

        # - すべて計測開始に失敗: start() 処理をエラーで終了
        if len(failed) == len(self._sensors):
            self._logger.error("All sensors failed to start measurement")
            raise LoggerError("All sensors failed to start measurement")

        # 計測情報ファイル 開始時出力: 計測終了時間を除き出力、終了時に上書き
        self._output_info()

        # 計測監視ループ: センサーサブプロセスからのエラーを定期的にチェック
        while self._measuring:
            time.sleep(1)
            # エラーキューに登録あり
            if self._error_queue and not self._error_queue.empty():
                error: JobError = self._error_queue.get()
                for sensor in self._sensors:
                    # エラー発生源センサー特定
                    if sensor.model == error.model and sensor.serial == error.serial:
                        failed.add(f"{error.model}_{error.serial}")
                        self._logger.error(
                            f"Sensor: {sensor} got error during measurement"
                        )
                        # センサー停止試行
                        try:
                            sensor.stop()
                        except Exception:
                            pass
                        break

                # 全てのセンサーの計測が失敗したらエラーで終了
                if len(failed) == len(self._sensors):
                    self._logger.error("All sensors failed to measure")
                    raise LoggerError("All sensors failed to measure")

        # センサー計測終了待ち： タイマーが設定されている場合、サブプロセスの終了を待つ
        if self._stop_timer:
            self._stop_timer.join()

    def stop(self, terminate: bool = False) -> None:
        """センサーによる計測を終了する

        このメソッドは、計測時間タイマーの発火か、main の終了シーケンスから呼ばれる

        Args:
            terminate (bool, optional): SIGTERM などの中断により呼ばれたか
                - デフォルト: False
        """
        # 計測中判定： main の終了シーケンスでは未計測時に呼ばれる場合がある
        if not self._measuring:
            self._logger.debug("stop method called, but not measure")
            return

        try:
            # 終了処理
            # - 内部計測フラグ OFF: start() での計測監視ループを停止する
            self._measuring = False

            # - 中断による終了時: 計測時間タイマーを停止する
            if terminate and self._stop_timer:
                self._stop_timer.cancel()
                self._stop_timer = None

            # - センサー停止指示
            for sensor in self._sensors:
                try:
                    sensor.stop()
                except Exception:
                    self._logger.exception(
                        f"Sensor: {sensor} failed to stop measurement"
                    )

            # - 終了日時記録
            self._end_time = datetime.datetime.now()

            # - 終了ログ
            if terminate:
                self._logger.info("Terminate measurement")
            else:
                self._logger.info("Finish measurement")

            # - 計測情報ファイル 終了時出力（上書き）
            self._output_info()

            # - エラーキューのクローズ
            if self._error_queue:
                self._error_queue.close()
                self._error_queue = None

        except Exception:
            # main の終了シーケンスで呼ばれるため、Error はログ出力して無視する
            self._logger.exception("An error occurred during stop measurement")

    def get_output_dir(self) -> Union[str, None]:
        """計測データの出力先フォルダを返す

        Returns:
            Union[str,None]: 計測が開始されている場合は、開始年月日を基準にした所定のパスを返す
                - 計測が開始されていない場合: None
        """
        return (
            f"{self.config.output_path}/{self._start_time:%Y%m%d_%H%M%S}"
            if self._start_time
            else None
        )

    @classmethod
    def _list_comports(cls) -> List[str]:
        """通信可能なポートの一覧を取得するメソッド

        Returns:
            List[str]: 通信できるポートの一覧
        """
        result = []
        for com in serial.tools.list_ports.comports():
            if com.device.startswith("/dev/ttyAMA"):
                # ラズパイの GPIO ポートのため使用しない
                continue
            result.append(com.device)
        return sorted(result)

    def _init_sensor(self, port: str) -> Union[Sensor, None]:
        """センサーの初期化をするメソッド

        Args:
            port (str): センサーと通信を試みるポート

        Raises:
            SensorDeviceError: センサーが通信可能な状態でない場合

        Returns:
            Union[Sensor, None]: 初期化に成功したセンサー
        """
        # センサーとの初期通信
        try:
            sensor = self._create_sensor(port)
        except ConnectionRefusedError:
            self._logger.error(f"Failed to connect: {port}")
            return None

        # センサーの状態チェック
        try:
            # 状態チェック
            if not sensor.ready():
                raise SensorDeviceError()

            # シリアル読み込み
            sensor.load_serial()
            # ファームウェアバージョン読み込み
            sensor.load_firm_ver()
        except Exception:
            self._logger.exception(f"Failed to initialize: {port}")
            return None

        return sensor

    def _create_sensor(self, port: str) -> Sensor:
        """センサー(具象クラス)を生成するメソッド

        Args:
            port (str): センサーとの通信を試みるポート

        Raises:
            ValueError: 不正なセンサーのproduct_idが取得された場合

        Returns:
            Sensor: 作成されたセンサー(具象クラス)
        """
        # センサークラスに PRODUCT_ID の取得を指示
        product_id = Sensor.load_product_id(port, self.config.baud_rate)
        self._logger.info(
            f"Found product: {product_id} port={port} baud={self.config.baud_rate}"
        )

        # A352 生成
        if product_id.startswith("A352"):
            return self._factory.create_A352(port, product_id)

        # A342 生成
        if product_id.startswith("A342"):
            return self._factory.create_A342(port, product_id)

        else:
            raise ValueError(f"Unknown PRODUCT_ID: {product_id}")

    def _output_info(self) -> None:
        """計測情報ファイルを出力する関数

        Raises:
            AssertionError: 計測開始前に呼ばれた場合（保存先ディレクトリが存在しない）
        """
        # 出力先フォルダ
        output_dir = self.get_output_dir()
        assert output_dir

        # 計測時間
        meas_time: Optional[int] = None
        if self._start_time and self._end_time:
            meas_time = math.floor((self._end_time - self._start_time).total_seconds())

        # センサーごとに出力
        for sensor in self._sensors:
            info_text = (
                f"{InfoKeys.OS},{platform.platform()}\n"
                + f"{InfoKeys.LOGGER_VERSION},{Constants.LOGGER_VERSION}\n"
                + f"{InfoKeys.LOGGER_ID},{self.config.logger_id}\n"
                + f"{InfoKeys.START_TYPE},{self._start_type}\n"
                + f"{InfoKeys.TIME_ZONE},{get_time_zone()}\n"
                + f"{InfoKeys.DATE},{app_format(self._start_time, '%Y/%m/%d %H:%M:%S')}\n"
                + f"{InfoKeys.MEASURE_TIME_SEC},{app_format(meas_time)}\n"
                + f"{InfoKeys.INITIAL_WAIT},{0 if self._start_type == Constants.START_TYPE_SCHEDULE else self.config.initial_wait}\n"
                + f"{InfoKeys.FILE_ROTATE_MIN},{self.config.file_rotate_min}\n"
                + f"{InfoKeys.BAUD_RATE},{self.config.baud_rate}\n"
                + f"{InfoKeys.PORT},{sensor.port}\n"
                + f"{InfoKeys.SENSOR},{sensor.model}\n"
                + f"{InfoKeys.PRODUCT_ID},{sensor.product_id}\n"
                + f"{InfoKeys.SERIAL_NO},{sensor.serial}\n"
                + f"{InfoKeys.FIRM_VER},{sensor.firm_ver:#06x}\n"
                + f"{InfoKeys.PHYSICAL},{sensor.physical}\n"
                + f"{InfoKeys.OUTPUT_TYPE},{sensor.mode}\n"
                + f"{InfoKeys.SPS},{sensor.sps}\n"
                + f"{InfoKeys.FILTER},{app_format(sensor.filter)}\n"
                + f"{InfoKeys.RMS_PP_INTERVAL},{app_format(sensor.rms_pp_interval)}\n"
                + f"{InfoKeys.SENSOR_DATA_DIAG},{sensor.sensor_data_diag}\n"
                + f"{InfoKeys.SENSOR_DATA_DIAG_SEC},{sensor.sensor_data_diag_sec:f}\n"
            )

            # 計測ディレクトリのinfo.csvに書き込み
            # - ファイルシステムに異常がある場合例外が発生するが、ファイル出力が失敗するため最初のセンサーで例外にする
            with open(
                f"{output_dir}/{sensor.model}_{sensor.serial}_info.csv",
                mode="w",
                encoding="UTF-8",
                newline="\r\n",
            ) as f:
                f.write(info_text)

    def self_test(self) -> None:
        """センサーのセルフテストを実行する"""

        self._logger.info("Start connection")

        # USBデバイスサーチ
        for port in Controller._list_comports():
            self._logger.debug(f"COM port found: {port}")
            # センサー初期化
            sensor = self._create_sensor(port)
            sensor.load_serial()
            self._sensors.append(sensor)
        if len(self._sensors) == 0:
            self._logger.error("No sensor found")
            raise LoggerError("No Ready Sensor found")

        # セルフテスト実行
        for sensor in self._sensors:
            sensor.self_test()
