# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

from multiprocessing import Queue
from multiprocessing.sharedctypes import Synchronized
from typing import TYPE_CHECKING

from logger.config import Config
from logger.measure.job import ReaderArgs, WriterArgs, reader_job, writer_job
from logger.process import ApplicationProcess

# - Sensor は型チェックでのみ使用： TYPE_CHECKING で循環参照を回避
if TYPE_CHECKING:
    from logger.measure.sensor import Sensor


class LoggerFactory:
    """
    ロガーシステム内で使われる主要な要素に対するファクトリクラス

    様々な適用先への対応を容易にする目的で、以下の要素の生成を拡張可能にする：
    - 具象 Sensor
    - Sensor Reader Process
    - Sensor Writer Process

    以下の手順で拡張対応を行う：
    1. このクラスのサブクラスを作成する
    2. サブクラスで拡張バージョンを返すファクトリメソッドを定義する
    3. main() 関数の Controller 生成時に作成したファクトリオブジェクトを渡す
    """

    def __init__(self, config: Config) -> None:
        self.config = config

    def create_A342(self, port: str, product_id: str) -> "Sensor":
        """
        A342 センサーのファクトリメソッド

        Args:
            port (str): センサーが接続されたポート
            product_id (str): センサーから取得した PRODUCT_ID

        Returns:
            A342 センサーオブジェクト
        """
        # - 循環参照を回避するためこのタイミングで import
        from logger.measure.A342 import A342

        return A342(
            port=port,
            baud=self.config.baud_rate,
            product_id=product_id,
            logger_id=self.config.logger_id,
            file_rotate_min=self.config.file_rotate_min,
            mode=self.config.A342_mode,
            physical=self.config.A342_physical,
            rms_pp_interval=self.config.A342_rmspp_output_interval,
            sensor_data_diag=self.config.sensor_data_diag,
            sensor_data_diag_sec=self.config.sensor_data_diag_sec,
        )

    def create_A352(self, port: str, product_id: str) -> "Sensor":
        """
        A352 センサーのファクトリメソッド

        Args:
            port (str): センサーが接続されたポート
            product_id (str): センサーから取得した PRODUCT_ID

        Returns:
            A352 センサーオブジェクト
        """
        # - 循環参照を回避するためこのタイミングで import
        from logger.measure.A352 import A352

        return A352(
            port=port,
            baud=self.config.baud_rate,
            product_id=product_id,
            sps=self.config.A352_sps,
            logger_id=self.config.logger_id,
            file_rotate_min=self.config.file_rotate_min,
            sensor_data_diag=self.config.sensor_data_diag,
            sensor_data_diag_sec=self.config.sensor_data_diag_sec,
        )

    def create_reader_process(
        self,
        queue: Queue,
        error_queue: Queue,
        reader_args: ReaderArgs,
        measuring: Synchronized,
    ) -> ApplicationProcess:
        """
        Sensor Reader プロセスのファクトリメソッド

        引数で受け取っているのはデフォルトの reader_job で使うもの。
        reader_job を置き換える場合には、必要に応じてこのメソッド内で変更してよい。

        Args:
            queue (Queue): 計測データを Writer に送信するキュー
            error_queue (Queue): エラー発生時にコントローラーに通知するキュー
            reader_args (ReaderArgs): 計測設定情報を保持するオブジェクト
            measuring (Synchronized): 計測中フラグ

        Returns:
            Reader 機能を持った ApplicationProcess オブジェクト
        """
        return ApplicationProcess(
            target=reader_job, args=(queue, error_queue, reader_args, measuring)
        )

    def create_writer_process(
        self, queue: Queue, error_queue: Queue, writer_args: WriterArgs, output_dir: str
    ) -> ApplicationProcess:
        """
        Sensor Writer プロセスのファクトリメソッド

        引数で受け取っているのはデフォルトの writer_job で使うもの。
        writer_job を置き換える場合には、必要に応じてこのメソッド内で変更してよい。

        Args:
            queue (Queue): 計測データを Reader から受信するキュー
            error_queue (Queue): エラー発生時にコントローラーに通知するキュー
            writer_args (WriterArgs): 計測設定情報を保持するオブジェクト
            output_dir (str): 計測情報の保存ディレクトリ

        Returns:
            Writer 機能を持った ApplicationProcess オブジェクト
        """
        return ApplicationProcess(
            target=writer_job, args=(queue, error_queue, writer_args, output_dir)
        )
