# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import datetime
import logging
import multiprocessing
from dataclasses import dataclass
from multiprocessing.sharedctypes import Synchronized
from typing import List, Optional, Tuple, Union

from logger.comm import Comm, CommandList
from logger.measure.data import MeasureData


@dataclass
class ReaderArgs:
    """Reader プロセスのジョブに渡す設定情報のデータクラス

    Sensor 継承クラスでは、このクラスを継承したクラスを実装し、to_reader_args() で生成する
    """

    model: str
    serial: str
    port: str
    baud: int
    start_command: CommandList
    end_command: CommandList
    record_size: int
    record_begin: int
    record_end: int
    record_per_sec: int
    count_diff: int
    count_max: int
    count_start: int
    sensor_data_diag: bool
    diag_broken_count: int
    read_length: int = Comm.DEFAULT_READ_BYTES
    timeout: float = Comm.DEFAULT_TIMEOUT

    def __post_init__(self) -> None:
        self._diag_broken_counts = {"x": 0, "y": 0, "z": 0}

    def format_packet(self, index: int, packet: List[int]) -> MeasureData:
        """1レコード分のデータを受け取って計測データに変換するメソッド

        ReaderArgsを継承するクラスで具体的な処理を記載する

        Args:
            index (int): 何番目のデータか
            packet (List[int]): 変換対象のデータ配列

        Raises:
            NotImplementedError: サブクラスで実装しなかった場合

        Returns:
            MeasureData: 変換後のpythonデータクラス
        """
        raise NotImplementedError

    def complement_data(
        self, prev: Optional[MeasureData], next: MeasureData
    ) -> List[MeasureData]:
        """カウント値をチェックしてデータを補完するメソッド

        Args:
            prev (Optional[MeasureData]): 補完済みの最後のデータ
            next (MeasureData): 補完前の新たに配列に追加するデータ

        Returns:
            List[MeasureData]: データ欠損を補完した計測データ配列
        """
        logger = logging.getLogger(
            f"{__name__}.{self.model}.{self.serial}.complement_data"
        )

        # 初回のみ： index=-1 のデータを prev とする
        if prev is None:
            prev = MeasureData(
                index=-1,
                count=self.count_start - self.count_diff,
                temperature=0,
                x=0,
                y=0,
                z=0,
                flag=0,
            )

        # 次のカウント値に一致するまで一回ごとの差分を足し続けて差分を計算する
        diff = self.count_diff
        while (diff + prev.count) % self.count_max != next.count:
            diff += self.count_diff

        # カウントの差分と想定される差分が同一の場合、データ損失がないためそのまま返す
        if diff == self.count_diff:
            next.index = prev.index + 1
            return [next]

        # 欠損データの補完
        logger.warning(
            f"Missing {(diff // self.count_diff) - 1} data from index: {prev.index + 1}. Complement them."
        )
        result: List[MeasureData] = []
        for _ in range(0, diff, self.count_diff):
            index = prev.index + 1
            count = (prev.count + self.count_diff) % self.count_max

            # - カウント が next と一致： ループの終端のため、元のデータを入れる
            if count == next.count:
                next.index = index
                result.append(next)
            # - データ補完
            else:
                complement = self._gen_invalid_data(index=index, count=count)
                result.append(complement)
                prev = complement

        return result

    def _gen_invalid_data(self, index: int, count: int) -> MeasureData:
        """データ抜けがあった際に不正な値であることを示すデータクラスオブジェクトを生成するメソッド

        サブクラスで実装すること

        Args:
            index (int): 不正な計測データのindex
            count (int): 不正な計測データのカウント値、complement_dataの中で前のデータから算出する

        Raises:
            NotImplementedError: サブクラスで実装しなかった場合

        Returns:
            MeasureData: 補完したindexとcountを持つ不正なことを表すデータクラスオブジェクト
        """
        raise NotImplementedError

    def diag_sensor_data(self, prev: Optional[MeasureData], next: MeasureData) -> None:
        """センサーデータのエラー判定"""

        logger = logging.getLogger(
            f"{__name__}.{self.model}.{self.serial}.diag_sensor_data"
        )
        if prev is None:
            return

        for axis in ["x", "y", "z"]:
            p = getattr(prev, axis)
            n = getattr(next, axis)
            if p == n:
                self._diag_broken_counts[axis] += 1
                if self._diag_broken_counts[axis] == self.diag_broken_count:
                    logger.warning(f"Sensor on axis: {axis} is possibly broken")
            else:
                if self._diag_broken_counts[axis] >= self.diag_broken_count:
                    logger.info(f"Sensor on axis: {axis} is fixed")
                self._diag_broken_counts[axis] = 0


@dataclass
class WriterArgs:
    """Writer プロセスのジョブに渡す設定情報のデータクラス

    Sensor 継承クラスでは、このクラスを継承したクラスを実装し、to_writer_args() で生成する
    """

    model: str
    serial: str
    logger_id: str
    port: str
    record_per_file: int

    def get_file_name(self, start_time: datetime.datetime) -> str:
        """与えられた時刻でデータを出力するcsvのファイル名を生成するメソッド

        Args:
            start_time (datetime.datetime): 出力開始時刻

        Returns:
            str: ディレクトリ名を含まないcsvファイルのファイル名
        """
        return f"{self.model}_{self.serial}_{self.logger_id}_{self.port[-1]}_{start_time:%y%m%d_%H%M%S}.csv"


def _packets_to_data(
    packets: List[int], args: ReaderArgs
) -> Tuple[List[MeasureData], List[int]]:
    """センサーから読み込んだデータから、計測データオブジェクトの配列に変換する関数

    Args:
        packets (List[int]): センサーから読み込んだデータ
        args (ReaderArgs): 計測設定情報を保持するオブジェクト

    Returns:
        Tuple[List[MeasureData],List[int]]: 変換済み計測データの配列, 1レコードに足りず変換できなかったデータ
    """
    logger = logging.getLogger(f"{__name__}.{args.model}.{args.serial}.packets_to_data")

    index = 0
    data: List[MeasureData] = []
    rest: List[int] = []
    while True:
        first_err = True
        packet = packets[index : index + args.record_size]

        # packetサイズが0になっている恐れもあるため、while文の条件として長さを判定する
        while len(packet) == args.record_size and (
            packet[0] != args.record_begin or packet[-1] != args.record_end
        ):
            if first_err:
                logger.warning(
                    f"Invalid packet boundary: beg={packet[0]:#X}, end={packet[-1]:#X}. Fix packet."
                )
                # 連続してログが出力されるのを抑止
                first_err = False

            # indexを1つずらして再調査
            index += 1
            packet = packets[index : index + args.record_size]

        # 切り取ったpacketが1レコード分に満たない時はデータの終端に来ているのでループを抜ける
        if len(packet) < args.record_size:
            rest = packet
            break

        data.append(args.format_packet(0, packet))

        # indexをレコード分だけずらして次のループへ
        index += args.record_size

    return data, rest


def reader_job(
    queue: multiprocessing.Queue,
    error_queue: multiprocessing.Queue,
    args: ReaderArgs,
    measuring: Synchronized,
) -> None:
    """Reader プロセス上でセンサーからデータを受け取り、整形してキューに渡すジョブ関数

    Args:
        queue (multiprocessing.Queue): 計測データを Writer に送信するキュー
        error_queue (multiprocessing.Queue): エラー発生時にコントローラーに通知するキュー
        args (ReaderArgs): 計測設定情報を保持するオブジェクト
        measuring (Synchronized): 計測中フラグ
    """
    logger = logging.getLogger(f"{__name__}.{args.model}.{args.serial}.reader")
    comm = Comm()
    try:
        # 計測開始
        comm.open(args.port, args.baud, args.timeout)
        comm.send(args.start_command)

        process_count = 0
        data: List[int] = []  # 未処理
        conv: List[MeasureData] = []  # 変換済
        send: List[MeasureData] = []  # 送信対象
        last: Optional[MeasureData] = None  # 補完済み最新

        # 計測中ループ
        while measuring.value:
            result = comm.read(args.read_length)
            data.extend(result)

            # パケットから計測データに変換する
            conv, rest = _packets_to_data(data, args)
            # - 未処理を更新
            data = rest

            # 変換データを補完しながら送信対象に詰めていく
            for cd in conv:
                if args.sensor_data_diag:
                    args.diag_sensor_data(last, cd)
                send.extend(args.complement_data(last, cd))
                last = send[-1]

            # 1秒以上のデータが溜まっていたら Writer プロセスに送出
            if len(send) >= args.record_per_sec:
                queue.put(send[0 : args.record_per_sec])
                # 送信対象を更新
                send = send[args.record_per_sec :]
                process_count += args.record_per_sec

        # 計測中フラグがオフになったら残りを送信する
        queue.put(send)

    except Exception as e:
        # サブプロセスでの例外発生は、ログ出力をしてエラーキューでコントローラーに送信する
        logger.exception("An error occurred during reading data")
        error_queue.put(JobError(model=args.model, serial=args.serial, error=e))

    finally:
        # Reader 終了シーケンス
        # - ここでは例外をキャッチしてログ出力はするが次へ進める
        try:
            # 終了コマンド
            comm.send(args.end_command)
        except Exception:
            logger.exception("Failed to send stop command")
        try:
            # 通信クローズ
            comm.close()
        except Exception:
            logger.exception("Failed to disconnect device connection")
        try:
            # Writer に終了を通知
            queue.put(None)
        except Exception:
            # - 終了通知に失敗すると Writer が残存する恐れがあるため FATAL
            logger.fatal("Failed to notify finishing writer process")


def writer_job(
    queue: multiprocessing.Queue,
    error_queue: multiprocessing.Queue,
    args: WriterArgs,
    output_dir: str,
) -> None:
    """Writer プロセス上で Reader プロセスから受け取ったデータをファイルに書き込むジョブ関数

    Args:
        queue (multiprocessing.Queue): 計測データを Reader から受信するキュー
        error_queue (multiprocessing.Queue): エラー発生時にコントローラーに通知するキュー
        args (WriterArgs): 計測設定情報を保持するオブジェクト
        output_dir (str): 計測情報の保存ディレクトリ
    """
    logger = logging.getLogger(f"{__name__}.{args.model}.{args.serial}.writer")

    # 書き込み設定
    mode = "w"
    write_count = 0
    output_file = f"{output_dir}/{args.get_file_name(datetime.datetime.now())}"
    logger.info(f"Initial output to {output_file}")

    try:
        # 出力ループ
        finish = False
        while True:
            # キューから取得
            # - Writer の処理間隔を広げるため、Reader 側で 1秒分のレコードをまとめて送出している
            records: Union[List[MeasureData], None] = queue.get()

            # - 終了判定： None の場合は計測終了
            if records is None:
                finish = True
                break

            # データ件数がある場合： 計測モードにより 0 レコード送信があるため回避
            if len(records) > 0:
                lines = "".join((d.to_csv_line() for d in records))
                with open(
                    output_file, mode=mode, encoding="UTF-8", newline="\r\n"
                ) as f:
                    f.write(lines)
                    logger.debug(f"write {len(records)} lines")
                write_count += len(records)

                # 以降は追記モード
                mode = "a"

            # ファイルローテート： ファイル名、モードを変更
            if write_count >= args.record_per_file:
                output_file = (
                    f"{output_dir}/{args.get_file_name(datetime.datetime.now())}"
                )
                mode = "w"
                write_count = 0
                logger.info(f"Output file is rotated to {output_file}")

    except Exception as e:
        # サブプロセスでの例外発生は、ログ出力をしてエラーキューでコントローラーに送信する
        logger.exception("An error occurred during writing data")
        error_queue.put(JobError(model=args.model, serial=args.serial, error=e))

    finally:
        # Writer 終了シーケンス
        if not finish:
            # 正常終了でない場合、Reader からの終了通知を受けるまでキューから取得して空にする
            try:
                while True:
                    # この段階では出力に失敗している可能性が高いため、終了通知以外は無視する
                    message: Union[List[MeasureData], None] = queue.get()
                    if message is None:
                        break
            except Exception:
                logger.fatal("Failed to get reader process finish message")


class JobError:
    """サブプロセスで発生したエラーについて、コントローラーでセンサーの特定が可能なように属性を付加したクラス"""

    def __init__(self, model: str, serial: str, error: Exception) -> None:
        self.model = model
        self.serial = serial
        self.error = error
