# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import multiprocessing
import signal
from typing import Any, Callable, Iterable, Mapping, Optional

from logger.log import LoggingClient, LoggingService


class ApplicationProcess(multiprocessing.Process):
    """アプリケーションに必要な処理を追加した multiprocessing.Process のラッパークラス

    以下の機能を持つ
    - プロセスセーフなファイル出力ロギング
    - キーボードインタラプトでプロセスを終了させない
    - Terminateでプロセスを終了させない
      - process.terminate() でプロセスが終了しなくなるため、必ず終了する関数を渡すこと

    Args:
        multiprocessing.Process に同じ
    """

    def __init__(
        self,
        group: None = None,
        target: Optional[Callable] = None,
        name: Optional[str] = None,
        args: Iterable[Any] = (),
        kwargs: Mapping[str, Any] = {},
        *,
        daemon: Optional[bool] = None,
    ) -> None:
        # ロギングオブジェクト取得
        log_client = LoggingService.get_instance().get_logging_client()
        super().__init__(
            target=ApplicationProcess._run_target,
            args=(log_client, target, args, kwargs),
            group=group,
            name=name,
            kwargs=kwargs,
            daemon=daemon,
        )

    @staticmethod
    def _run_target(
        log_client: LoggingClient,
        target: Optional[Callable] = None,
        args: Iterable[Any] = (),
        kwargs: Mapping[str, Any] = {},
    ):
        """指定された関数をサブプロセス上で実行させる関数"""

        # シグナル処理
        # - サブプロセスでは SIGTERM, SIGINT を無視する
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        # マルチプロセスロギング構築
        log_client.configure(log_client.queue)

        # 指定処理の実行
        if target:
            target(*args, **kwargs)


def prepare_process() -> None:
    """サブプロセスの生成方法をロガープログラムに適したものに変更する

    この関数を main 関数の冒頭で呼び出すこと
    """
    multiprocessing.set_start_method("spawn")
