# Copyright 2024, Seiko Epson Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the “Software”), to deal in
# the Software without restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
# Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.

import logging
import signal
import sys
import time

import schedule

from logger.constants import ConfigKeys, Constants, ExitCode
from logger.error import TerminateException, raise_terminate_exception
from logger.log import init_stdout_logger
from logger.tool.hwmonitor.manager import MonitorManager
from logger.utils.env import get_str
from logger.utils.validator import Validator


def main():
    """メイン関数 for hwmonitor"""

    # プロセス初期化
    # - シグナルハンドラー登録 : SIGTERM, SIGINT 発生時は例外処理にて対処する
    signal.signal(signal.SIGTERM, raise_terminate_exception)
    signal.signal(signal.SIGINT, raise_terminate_exception)

    # - ロギング初期化
    LOG_LEVEL: str = str(
        get_str(ConfigKeys.LOG_LEVEL)
        if get_str(ConfigKeys.LOG_LEVEL) in Constants.LOG_LEVELS
        else Constants.LOG_LEVEL_DEFAULT
    )
    init_stdout_logger(level=LOG_LEVEL)
    logger = logging.getLogger(__name__)

    # 監視開始
    logger.info("Start monitor")
    try:
        # - 保存先フォルダチェック
        validator = Validator()
        output_path = get_str(ConfigKeys.OUTPUT_PATH)
        result = validator.exist_dir(output_path, ConfigKeys.OUTPUT_PATH)
        if not result:
            logger.error("Can't monitor")
            return ExitCode.EX_NG

        # - 監視初期化
        manager = MonitorManager(output_path=str(output_path))
        manager.init()

        # - 監視処理関数
        def job() -> None:
            result = manager.monitor()
            logger.info(result)
            manager.output(result)

        # - １分間隔で処理を実行
        schedule.every(1).minutes.do(job)

        # 監視実行
        while True:
            schedule.run_pending()
            time.sleep(1)

    except TerminateException:
        logger.info("Finish monitor")
        return ExitCode.EX_OK
    except Exception:
        logger.exception("Unknown error occurred")
        return ExitCode.EX_NG


if __name__ == "__main__":
    code = main()
    sys.exit(code)
